"""
This module initializes the environment for koatl.
It sets up the meta-finder hook to enable importing .tl files,
and also declares functions that are required for certain tl features to work,
such as coalescing and module exports.

koatl.runtime should be written in Python only since otherwise
it would create a circular dependency.
"""

import functools
from types import SimpleNamespace

from . import meta_finder

meta_finder.install_hook()


from .virtual import *
from .record import *
from .helpers import *
from .classes import *


def dummy(name):
    def wrapper(*args, **kwargs):
        raise RuntimeError(
            f"{name} is not available without the prelude. Please import koatl.prelude."
        )

    return wrapper


__tl__ = SimpleNamespace(
    Exception=Exception,
    slice=slice,
    type=type,
    vget=virtual.vget,
    vhas=virtual.vhas,
    set_exports=helpers.set_exports,
    do=helpers.do,
    partial=functools.partial,
    # These require more complex logic and require the prelude.
    # The runtime provides dummy implementations that raise if used without the prelude.
    memo_value=dummy("memo"),
    async_memo_value=dummy("memo"),
    op_map=dummy("?"),
    op_coal=("??"),
    Ok=dummy("try-expr"),
    Err=dummy("try-expr"),
    Result=dummy("Result"),
    **{name: helpers.__dict__[name] for name in helpers.__all__},
    **{name: record.__dict__[name] for name in record.__all__},
    **{name: virtual.__dict__[name] for name in virtual.__all__},
    **{name: classes.__dict__[name] for name in classes.__all__},
)


__all__ = [
    "__tl__",
    *helpers.__all__,
    *record.__all__,
    *virtual.__all__,
    *classes.__all__,
]



from functools import partial
from itertools import count
from types import SimpleNamespace
from .._rs import fast_vget, fast_vset, fast_vset_trait


def vget(obj, name, ignore_traits=False):
    try:
        return getattr(obj, name)
    except AttributeError:
        pass

    # special case for iter - this could be implemented using types and trait vtbls
    # but this is simpler and probably faster
    if name == "iter":
        if isinstance(obj, slice):
            start = obj.start if obj.start is not None else 0
            step = obj.step if obj.step is not None else 1
            if obj.stop is None:
                return iter(count(start, step))
            else:
                return iter(range(start, obj.stop, step))

        try:
            return iter(obj.items())
        except AttributeError:
            pass

        try:
            return iter(obj)
        except TypeError:
            pass

    v = fast_vget(obj, name, ignore_traits)
    if v is not None:
        if hasattr(v, "_property"):
            return v(obj)

        return partial(v, obj)

    raise AttributeError(f"'{type(obj).__name__}' object has no v-attribute '{name}'")


def vhas(obj, name, ignore_traits=False):
    if hasattr(obj, name):
        return True

    if name == "iter":
        if isinstance(obj, slice):
            return True

        try:
            obj.items()
            return True
        except AttributeError:
            pass

        try:
            iter(obj)
            return True
        except TypeError:
            pass

    v = fast_vget(obj, name, ignore_traits)
    if v is not None:
        return True

    return False


def ext_prop(type, name):
    def impl(value):
        value._property = True
        fast_vset(type, name, value)
        return value

    return impl


def ext_method(type, name):
    def impl(value):
        fast_vset(type, name, value)
        return value

    return impl


def ext_trait(type):
    for name, value in type._own_methods.items():
        fast_vset_trait(type.__name__, type._trait_reqs, name, value)
    return type


Extension = SimpleNamespace(
    property=ext_prop,
    method=ext_method,
    trait=ext_trait,
)


__all__ = [
    "Extension",
]
